"""This module provides support for discrete Fourier transforms.

It calculates the discrete Fourier transform using:

   X(k) = \sum_{n=0}^{N-1} x(n) e^{-j * 2 * \pi * n * k / N}

Copyright 2020--2021 Michael Hayes, UCECE
Juergen Weizenecker, HKA
"""

import sympy as sym
from .transformer import BilateralForwardTransformer
from .sym import sympify, AppliedUndef, j, pi, miscsymbol
from .extrafunctions import UnitImpulse, UnitStep
from .utils import factor_const, scale_shift
from .matrix import Matrix
from .ztransform import is_multiplied_with
from copy import deepcopy
from warnings import warn


__all__ = ('DFT', 'DFTmatrix')


def is_in_interval(n0, n1, nk, comment=''):
    """Check if bin belongs to an interval [n0, n1].
     n0, n1, nk may contain parameter

    This returns:
     -1  if bin is not in interval and is on the left side
      0  if bin is in interval
      1  if bin is not in interval and is on the right side

    If n0, n1 or nk contain parameter, 0 is returned if no decision is
    possible and a comment is printed.

    """

    # Is in [n0, n1]
    if sym.sympify(nk - n0).is_nonnegative and sym.sympify(n1 - nk).is_nonnegative:
        return 0
    # Is on the left
    elif sym.sympify(nk - n0).is_negative:
        return -1
    # Is on the right
    elif sym.sympify(n1 - nk).is_negative:
        return 1
    # No decision, assume in interval
    else:
        warn('DFT of %s, assuming %s is in interval [%s, %s]' %
             (comment, nk, n0, n1))
        return 0


def find_pow(q_expr, base, NN):
    """This finds terms of the form base**NN in the expression q_expr.
    It is used for simplifying q**N terms to one if q=exp(-+j*2*pi/N*k).

    It returns a list of all bases found e.g., [q/3, q/5, 2q, aq,  .......].
    """

    result = []
    for x in q_expr.atoms(sym.Pow):
        this_ex = x.as_base_exp()[1]
        this_ba = x.as_base_exp()[0]
        if this_ex == NN and not sym.simplify(this_ba / base).has(base):
            result += [this_ba]

    return result


def special_diff(var, Q, o, s, var_a, var_N):
    """Make derivative of a**(k-s)*Q / (1-a**N)**o with respect to a, and
       divide result by a**(k-s-1)/(1-a**N)**(o+1) to obtain a
       polynom in n.  This is used with make_polynom_table.

    """
    return ((var - s) * Q + var_a*sym.diff(Q, var_a)) * (1 - var_a**var_N) + o * var_N * Q * var_a**var_N


def make_polynom_table1(var, MaxP, aa=None, NN=None, prn=False):
    """ This Table contains solutions of the DFT of
          exp(j*2*pi/N*n*mu) / (1-a*exp(j*2*pi/N*n))**p
          where mu < p  and mu, p are integer
          NO roots in the denominator

          use the table as follows
          the list Qpoly[p] contains polynoms with respect to k of the solutions to a specific p and  0<=mu<=p-1 .
          The full solution is obtained as
          N * a**(k-mu) / (1-a**N)**p / (p-1) *Qpoly[p][mu]           p >= 2
          N*a**k / (1-a**N)* TP[1]                                                       p =1

           Example for p=3
           1 / (1-a*exp(-j*2*pi/N*n))**3  = N*a**k * Qpoly[3][0] / (1-a**N)**3 / 2
           exp(-j*2*pi/N*n) / (1-a*exp(-j*2*pi/N*n))**3  = N*a**(k-1) * Qpoly[3][1] / (1-a**N)**3 / 2
           exp(-j*2*pi/N*n*2) / (1-a*exp(-j*2*pi/N*n))**3  = N*a**(k-2) * Qpoly[3][2] / (1-a**N)**3 / 2

          The results for all p's are calculated (--> Qpoly), only p=MaxP is returned.
          The parameter a and N can be set, if not a and N remain a symbol
    """
    a = sym.Symbol('a')
    N = sym.Symbol('N', integer=True)
    Qpoly = [[], [1]]
    # Make the Polynoms Q_p_mu
    # Run through all p, as the formula is recursiv (manuscript)
    for l in range(2, MaxP + 1, 1):
        # Special derivative when polynoms are involved (see also handle n**p in termXq)
        Q0 = special_diff(var, Qpoly[l - 1][0], l - 1, 0,
                          a, N) + Qpoly[l - 1][0] * (1 - a**N) * (l - 1)
        this_Qpoly = [sym.simplify(sym.expand(Q0))]
        # Run for each p through all 0<=mu<=p-1
        for m in range(l - 1):
            Qm = special_diff(var, Qpoly[l - 1][m], l - 1, m, a, N)
            this_Qpoly += [sym.simplify(sym.expand(Qm))]

        # Write result to a  list containing all polynomials
        Qpoly += [this_Qpoly]

    # Return polynoms to p=MaxP with specific a and N given
    ret_pol = []
    for mu in range(MaxP):
        pol = 0
        for r in range(MaxP):
            dj = sym.expand(Qpoly[MaxP][mu]).coeff(var, r)
            if not aa is None:
                dj = dj.subs(a, aa)
            if not NN is None:
                dj = dj.subs(N, NN)
            pol += dj * var**r
            if prn:
                print("mu = ", mu, " kn**", r, "  :  ", dj)
        if prn:
            print("Q = ", pol, "\n")
        ret_pol += [pol]
    # Return polynom for MaxP only, otherwise Qpoly must be returned
    return ret_pol


def make_polynom_table2(var, p, varN=None, prn=False):
    """This Table contains solutions of the DFT of
        exp(j*2*pi/N*n*mu) / (1-a*exp(j*2*pi/N*n))**p
        where mu < p  and mu, p are integer
        a is of the form exp(-j*2*pi*r/N), so here we have roots in the denominator at n=r

        The solution is of the form
            Q[mu] * exp(-j*2*pi*r/N*(k-mu)), were Q[mu] is a polynom in k of order p

            idea q = exp(j*2*pi/N*(n-r)):
           The transform of C_p(q) / (1-q)**p is (k**p - 1/N*sum_0_N-1(k**p)) * exp(-j*2*pi*mu*k/N)                 (1)

           Find from these the transforms of 1/(1-q)**p, q/(1-q)**p, ...., q**(p-1)/(1-q)**p by an appropriate
           linear combination of the transforms in (1).
           The C_p(q) is the polynom according to  sum_0_(N-1) n**p * q**n = N*Cp / (1-q)**p with q**N=1
    """
    NN = sym.Symbol('NN')
    q = sym.Symbol('q')
    a_p = 1 + 0 * q
    b_p = 1 + 0 * q
    AllC = [0]
    # Find first the C_p with the recursive formula for the repetitive derivative   q*diff((1-q**N)/(1-q), q)
    # by writing this as (a_p - q**N*b_p) /(1-q)**p.
    for i in range(p):
        a_p = sym.expand((q - q**2) * sym.diff(a_p, q) + (i + 1) * q * a_p)
        b_p = sym.expand(NN * (1 - q) * b_p + (q - q ** 2) *
                         sym.diff(b_p, q) + (i + 1) * q * b_p)
        pol, rem = sym.div(a_p - b_p, 1 - q)
        AllC += [(pol/NN).as_poly(q).as_expr()]
    if prn:
        print("p=1, ...., ", p, "  Ci(q) = ", AllC[1:])

    # Solve the equation x_p*C_p + x_(p-1)*C_(p-1)(1-q) +... = q**s = r_1 +r_2*q + .. r_p*q**(p-1)
    # to find the linear factors xi by equating coefficients in the variable q
    x = [sym.symbols('x%d' % i) for i in range(p)]
    r = [sym.symbols('r%d' % i) for i in range(p)]
    pol = 0 * q
    for i in range(p):
        # Left side of equation
        pol += x[i] * AllC[i + 1] * (1 - q) ** (p - i - 1)

    # Expand left side in orders of q
    coeffs = sym.Poly(sym.expand(pol), q).all_coeffs()[-1::-1]
    # print("Left side equation ", coeffs)
    eq = []
    xi = []
    # Subtract right hand side
    for i in range(p):
        eq += [coeffs[i] - r[i]]
        xi += [x[i]]

    # Solve for the xi by equating coefficients in q
    sol = sym.linsolve(eq, tuple(xi))
    All_rhs = sym.eye(p)
    # Set r_1, .., r_p to obtain the solution for all desired new transforms
    # e.g.  if q**2/(1-q)**p is to be found, then chose r_1=r_2=0    r_3=1   r_4=..=r_p=0
    ret_pol = []
    for i in range(p):
        rhs = list(All_rhs[:, i])
        repl = dict(zip(r, rhs))

        if prn:
            print(
                "q**", i, " :  linear factors of basic transform [x_i] = ", sol.subs(repl).args[0])
        # Find finally the polynoms in k for the new transforms  1/(1-q), , , , q**(p-1)/(1-q)**p
        trans_i = 0 * var
        for ii in range(p):
            # Sum k**(ii+1)
            faulh = sym.factor(
                (sym.bernoulli(ii + 2, NN) - sym.bernoulli(ii + 2, 0)) / (ii + 2))
            trans_i += sol.subs(repl).args[0][ii] * \
                (var ** (ii + 1) - faulh / NN)
        if not varN is None:
            trans_i = trans_i.subs(NN, varN)
        ret_pol += [trans_i.as_poly(var).as_expr()]
        if prn:
            print("q**", i, " :  D(.) = ", trans_i.as_poly(var).as_expr(), "\n")

    return ret_pol


# Does not use simplify as sometimes weird expression may be generated
# e.g. exp(j*2*pi/N) = (-1)**(2/N), or denoms are expanded and not
# factorized or it takes to long
def simp_rat(expr, q, method=1):
    """
    Make a few simplification steps for a rational function, useful for DFT's
    method 1: uses cancel
    method 2: uses factor_list of num and denom and cancel their common factors by hand
    """
    if method == 1:
        #ret_expr = sym.simplify(expr)
        #ret_expr = sym.cancel(expr, q)
        try:
            # that does not work with parameter
            ret_expr = sym.cancel(expr, q)
        except:
            ret_expr = sym.cancel(expr)
            return ret_expr
        Denom = sym.denom(ret_expr)
        # Denominator was probably expanded, so return original one
        if Denom.is_Add and len(Denom.args) == 3:
            return sym.numer(expr).as_poly(q).as_expr() / sym.denom(expr)
        else:
            ret_expr = sym.numer(ret_expr).as_poly(
                q).as_expr() / sym.factor(Denom, q)
            return ret_expr
    elif method == 2:
        Num = sym.numer(expr)
        Denom = sym.denom(expr)
        # Make factor list
        f_n = list(sym.factor_list(Num, q))
        f_d = list(sym.factor_list(Denom, q))
        # Check for common factors
        for ii, xx in enumerate(f_d[1]):
            for jj, yy in enumerate(f_n[1]):
                hits = sym.cancel(yy[0] / xx[0])
                if not hits.has(q):
                    # Common factor found
                    f_n[0] *= hits**yy[1]
                    # Exponents
                    p_n = yy[1]
                    p_d = xx[1]
                    # Cancel common factors and write as positive exponent
                    if p_n <= p_d:
                        f_n[1][jj] = (xx[0], 0)
                        f_d[1][ii] = (xx[0], p_d-p_n)
                    else:
                        f_n[1][jj] = (xx[0], p_n-p_d)
                        f_d[1][ii] = (xx[0], 0)
                    break

        # Rebuild factorization
        Denom = 1 + 0 * q
        Num = f_n[0] / f_d[0]
        for xx in f_d[1]:
            Denom *= xx[0]**xx[1]
        for xx in f_n[1]:
            Num *= xx[0]**xx[1]
        ret_expr = Num/Denom
        return ret_expr
    else:
        raise ValueError('Invalid method')


class QkTransform(object):
    """This handles X(q) transforms, where

    X(q) = sum_lower^upper x(n) * q**N

    As special cases are possible (e.g., for q=0) the class handles the
    general expression and special cases"""

    def __init__(self, val_q, k0=None, val0=None, valqk=None, val_k=0):
        self.Xq = val_q
        self.Xk = val_k
        if k0 is not None:
            self.has_special = True
            self.cases = [k0]
            self.case_expr = {k0: [val0, valqk]}
        else:
            self.has_special = False
            self.cases = []
            self.case_expr = {}

    def multiply(self, c0):
        """Multiply by constant."""

        self.Xq *= c0
        self.Xk *= c0
        if self.has_special:
            for key in self.case_expr:
                self.case_expr[key][0] *= c0
                self.case_expr[key][1] *= c0

    def rm_cases(self):
        """Delete special cases."""

        if self.has_special:
            self.has_special = False
            self.cases = []
            self.case_expr = {}

    def shift_k(self, dk):
        """Shift the special cases k values by dk."""

        if self.has_special:
            self.cases = [ii + dk for ii in self.cases]
            new_dict = {}
            for key in self.case_expr:
                new_dict[key + dk] = self.case_expr[key]
            self.case_expr = new_dict

    def add(self, sec):
        """Merge (add) two QkTransforms."""

        self.Xq += sec.Xq
        self.Xk += sec.Xk
        # Both have special cases
        if sec.has_special and self.has_special:
            self.cases += [ii for ii in sec.cases if ii not in self.cases]
            for key in sec.case_expr:
                if key in self.case_expr:
                    self.case_expr[key][0] += sec.case_expr[key][0]
                    self.case_expr[key][1] += sec.case_expr[key][1]
                else:
                    self.case_expr[key] = sec.case_expr[key]
        # Only the second has special cases
        elif sec.has_special and not self.has_special:
            self.has_special = True
            self.cases = [ii for ii in sec.cases]
            for key in sec.case_expr:
                self.case_expr[key] = sec.case_expr[key]

    def subs(self, q, q_new):
        """Substitute q by a new expression."""

        self.Xq = (self.Xq).subs(q, q_new)
        if self.has_special:
            for key in self.case_expr:
                self.case_expr[key][1] = self.case_expr[key][1].subs(q, q_new)

    def simp_qN(self, ba, NN, all=False):
        """Simplifies ba**NN to 1

        If all is False, only the special cases will regarded.
        Replace e.g., (q/3)**NN by (1/3)**N
        """

        # Simplify for "no special cases" or if demanded
        if all or not self.has_special:
            # General expresion
            terms = find_pow(self.Xq, ba, NN)
            for full_base in terms:
                self.Xq = self.Xq.replace(full_base**NN, (full_base / ba)**NN)

        # Special cases
        if self.has_special:
            for key in self.case_expr:
                terms = find_pow(self.case_expr[key][1], ba, NN)
                for full_base in terms:
                    self.case_expr[key][1] = self.case_expr[key][1].replace(
                        full_base**NN, (full_base / ba)**NN)

    def make_transform(self, NN, backward, q, sub, k, piecewise):
        """Assemble the final tranform from all the cases, the general case, or Xk.

        backward is True for IDFT
        q is replaced by sub
        """
        simp_method = 1  # use cancel
        # simp_method = 2  # use home made factorization with cancellation
        if self.has_special:
            for ca in self.cases:
                k0 = ca
                # Shift k0 to [0, N-1]
                if backward and sym.sympify(ca).is_positive:
                    k0 -= NN
                elif not backward and sym.sympify(ca).is_negative:
                    k0 += NN

                arg_delta = k - k0
                if backward:
                    # as k=-n and k0 is negative
                    arg_delta *= -1

                # Special case * delta
                self.Xk += self.case_expr[ca][0] * UnitImpulse(arg_delta)

                # Rest * (1-delta) to avoid unwanted superposition try
                # to make expressions as compact as possible by using
                # simp_rat
                this_expr = self.case_expr[ca][1]
                if this_expr.is_Add:
                    sum_expr = 0 * q
                    for expr_i in this_expr.args:
                        sum_expr += simp_rat(expr_i, q, method=simp_method)
                    this_expr = sum_expr
                else:
                    this_expr = simp_rat(this_expr, q, method=simp_method)
                this_expr = this_expr.subs(q, sub)

                # Check if val is at k0 is unlimted and prepare for piecewise
                if not backward:
                    case_val = this_expr.subs(k, k0)
                    neq = sym.Ne(k, k0)
                else:
                    case_val = this_expr.subs(-k, -k0)
                    neq = sym.Ne(-k, -k0)

                if (case_val == sym.nan or case_val.has(sym.zoo) or
                        case_val == sym.oo or case_val == -sym.oo):

                    if piecewise:
                        self.Xk += sym.Piecewise((this_expr, neq))
                    else:
                        self.Xk += this_expr * (1 - UnitImpulse(arg_delta))
                else:
                    # Finite value, so write slightly differently
                    self.Xk += this_expr - case_val*UnitImpulse(arg_delta)
        else:
            this_expr = self.Xq
            if this_expr.is_Add:
                sum_expr = 0 * q
                for expr_i in this_expr.args:
                    sum_expr += simp_rat(expr_i, q, method=simp_method)
                this_expr = sum_expr
            else:
                this_expr = simp_rat(this_expr, q, method=simp_method)
            self.Xk = this_expr.subs(q, sub) + self.Xk


class DFTTransformer(BilateralForwardTransformer):

    name = 'DFT'
    is_inverse = False

    def key(self, expr, n, k, **kwargs):
        return expr, n, k, kwargs.get('N', None), kwargs.get('piecewise', False)

    def noevaluate(self, expr, n, k):

        foo = expr * sym.exp(-2 * j * pi * n * k / self.N)
        result = sym.Sum(foo, (n, 0, self.N - 1))
        return result

    def check(self, expr, n, k, N, **kwargs):

        try:
            N = N.expr
        except:
            pass

        # Convert constants to SymPy expressions.
        N = sym.sympify(N)

        if not N.is_integer and not N.is_positive:
            raise ValueError(
                "%s not positive integer, redefine as %s = symbol('%s', integer=True, positive=True)" % (N, N, N))

        symbols = expr.free_symbols
        symbol_names = [str(e) for e in symbols]
        if str(N) in symbol_names and N not in symbols:
            warn('There is a symbol in the expression with the same name as the DFT size %s but is not the same symbol' % N)

        self.N = N

        if expr.has(k):
            self.error('Expression depends on k')

        if expr.is_Piecewise and expr.args[0].args[1].has(n >= 0):
            self.error('Expression is unknown for n < 0 (use causal=True)')

    def sympy(self, expr, n, k):

        foo = expr * sym.exp(-2 * j * pi * n * k / self.N)
        result = sym.summation(foo, (n, 0, self.N - 1))

        return result

    def func(self, expr, n, k):

        if not isinstance(expr, AppliedUndef):
            self.error('Expecting function')

        scale, shift = scale_shift(expr.args[0], n)

        # Convert v(n) to V(k), etc.
        name = expr.func.__name__
        if self.is_inverse:
            func = sym.Function(name[0].lower() + name[1:])
        else:
            func = sym.Function(name[0].upper() + name[1:])

        result = func(k / scale) / abs(scale)

        if shift != 0:
            if self.is_inverse:
                shift = -shift
            result = result * \
                sym.exp(2 * sym.I * sym.pi * k * shift / (scale * self.N))

        if self.is_inverse:
            result *= self.N

        return result

    def function(self, expr, n, k):

        # Handle expressions with a function of FOO, e.g.,
        # v(n), v(n) * y(n), 3 * v(n) / n, v(4 * a * n), etc.,

        if not expr.has(AppliedUndef):
            self.error()

        const, expr = factor_const(expr, n)

        if isinstance(expr, AppliedUndef):
            return self.func(expr, n, k) * const

        tsym = sympify(str(n))
        expr = expr.subs(tsym, n)

        rest = sym.S.One
        undefs = []
        for factor in expr.as_ordered_factors():
            if isinstance(factor, AppliedUndef):
                if factor.args[0] != n:
                    self.error('Weird function %s not of %s' % (factor, n))
                undefs.append(factor)
            else:
                rest *= factor

        if rest.has(AppliedUndef):
            # Have something like 1/v(n)
            self.error()

        exprs = undefs
        if rest.has(n):
            exprs = exprs + [rest]
            rest = sym.S.One

        result = self.term(exprs[0], n, k) * rest

        if len(exprs) == 1:
            return result * const

        dummy = 'm' if self.is_inverse else 'l'

        for m in range(len(exprs) - 1):
            if m == 0:
                nu = miscsymbol(dummy, integer=True)
            else:
                nu = miscsymbol(dummy + '_%d' % m, integer=True)
            expr2 = self.term(exprs[m + 1], n, k)
            # Should be a circular convolution.
            result = sym.Sum(result.subs(k, k - nu) * expr2.subs(k, nu),
                             (nu, 0, self.N - 1)) / self.N

        return result * const

    # Make transform  Xq = sum_lower^upper  x[n] * q**n
    def termXq(self, expr, n, k, q, lower, upper):

        const, expr = factor_const(expr, n)
        args = expr.args
        xn_fac = []

        # Check for constant.
        if not expr.has(n):
            result_q = const * expr * q**lower * \
                (1 - q**(upper - lower + 1)) / (1 - q)
            # Special case k = 0
            result_1 = const * expr * (upper - lower + 1)
            result = QkTransform(result_q, 0, result_1, result_q)
            return result

        # Handle delta(n-n0)
        elif (expr.is_Function and expr.func == UnitImpulse and
              ((expr.args[0]).as_poly(n)).is_linear):
            aa = args[0].coeff(n, 1)
            bb = args[0].coeff(n, 0)
            nn0 = -bb / aa
            if nn0.is_integer:
                if is_in_interval(lower, upper, nn0, comment='delta') == 0:
                    # Shift frequency to -pi/2 ...  pi/2
                    if nn0.has(self.N):
                        nn0 = nn0.subs(self.N, 0)
                    elif sym.sympify(nn0 - self.N / 2).is_positive:
                        nn0 -= self.N

                    result_q = const * q**nn0
                    result = QkTransform(result_q)
                else:
                    result = QkTransform(0 * q)
                return result
            else:
                print("delta: bin ", nn0, " is not an integer")

        # Handle n**p
        if (expr == n or
            (expr.is_Pow and args[1].is_integer and args[1].is_positive
             and args[0] == n)):
            p = 1
            try:
                p = args[1]
            except:
                pass
            # Derivatives
            result_q = (q**lower - q**(upper + 1)) / (1 - q)
            # Make the first few by hand to make expressions shorter
            # and faster, use in all cases for the derivatives (and
            # multiplication with q) the structure
            # (q**lower * polynomial (q)- q**(upper+1)*polynomial(q)) (1-q)**(p+1) . The order of the polynomials is equal to p
            if p == 1:
                A_l = lower + q * (1 - lower)
                B_u = -q * upper + upper + 1
            elif p == 2:
                A_l = lower**2 + q**2 * \
                    (lower**2 - 2 * lower + 1) + q * \
                    (-2 * lower**2 + 2 * lower + 1)
                B_u = q**2*upper**2 + q * \
                    (-2 * upper**2 - 2 * upper + 1) + upper**2 + 2 * upper + 1
            elif p == 3:
                A_l = lower**3 + q**3 * (-lower**3 + 3 * lower**2 - 3 * lower + 1) + q**2*(
                    3 * lower**3 - 6*lower**2 + 4) + q*(-3 * lower**3 + 3 * lower**2 + 3 * lower + 1)
                B_u = -q**3 * upper**3 + q**2*(3 * upper**3 + 3 * upper**2 - 3 * upper + 1) + q*(
                    -3 * upper**3 - 6*upper**2 + 4) + upper**3 + 3 * upper**2 + 3 * upper + 1
            # Generate the polynoms A_l, B_u individually
            # TODO additional option to make a smart simplification by inspection of upper and lower
            else:
                # First polynoms
                A_l = 1 + 0 * q
                B_u = 1 + 0 * q
                # Find the next recursively
                for i in range(p):
                    A_l = sym.expand(
                        lower * A_l + (q - q**2) * sym.diff(A_l, q) - q * lower * A_l + (i + 1) * q * A_l)
                    B_u = sym.expand((upper + 1) * B_u + (q - q**2) * sym.diff(
                        B_u, q) - (upper + 1) * B_u * q + (i + 1) * q * B_u)
                A_l = A_l.as_poly(q).as_expr()
                B_u = B_u.as_poly(q).as_expr()
            result_q = const * (q**lower * A_l - q **
                                (upper + 1) * B_u) / (1 - q)**(p + 1)

            # Special case k=0, use Faulhaber's formula
            result_1 = const * \
                sym.factor((sym.bernoulli(p + 1, upper + 1) -
                           sym.bernoulli(p + 1, lower)) / (p + 1))
            result = QkTransform(result_q, 0, result_1, result_q)
            return result

        # Handle * rect((n-a)/b)
        elif is_multiplied_with(expr, n, 'rect', xn_fac):
            expr /= xn_fac[-1]
            ref = xn_fac[-1].args
            bb = 1 / sym.expand(ref[0]).coeff(n, 1)
            aa = -bb * sym.expand(ref[0]).coeff(n, 0)
            # Left and right index
            nn0 = aa - bb // 2
            nn1 = nn0 + bb - 1
            if (not aa.is_integer) or (not bb.is_integer):
                print("rect((n-a)/b): parameter not an integer")
            else:
                r1 = is_in_interval(lower, upper, nn0,
                                    comment='rect left step')
                r2 = is_in_interval(lower, upper, nn1,
                                    comment='rect right step')
                if r1 == 0 and r2 == 0:
                    result = self.termXq(expr, n, k, q, nn0, nn1)
                elif r1 == -1 and r2 == 0:
                    result = self.termXq(expr, n, k, q, lower, nn1)
                elif r1 == 0 and r2 == 1:
                    result = self.termXq(expr, n, k, q, nn0, upper)
                elif r1 == -1 and r2 == 1:
                    result = self.termXq(expr, n, k, q, lower, upper)
                else:
                    result = QkTransform(0 * q)

                result.multiply(const)
                return result

        # Handle * u(n-n0)
        elif is_multiplied_with(expr, n, 'UnitStep', xn_fac):
            expr /= xn_fac[-1]
            ref = xn_fac[-1].args
            aa = ref[0].coeff(n, 1)
            bb = ref[0].coeff(n, 0)
            if abs(aa) != 1:
                print("Use u(n-n0)")
            if not bb.is_integer:
                print("Step(n-n0): n0 not an integer")
            else:
                # Positive step
                if aa == 1:
                    nn0 = -bb
                    rg = is_in_interval(lower, upper, nn0, comment='step')
                    if rg == -1:
                        result = self.termXq(expr, n, k, q, lower, upper)
                    elif rg == 0:
                        result = self.termXq(expr, n, k, q, nn0, upper)
                    else:
                        result = QkTransform(0 * q)

                # Negative step
                elif aa == -1:
                    nn0 = bb
                    rg = is_in_interval(lower, upper, nn0, comment='Step:')
                    if rg == 1:
                        result = self.termXq(expr, n, k, q, lower, upper)
                    elif rg == 0:
                        result = self.termXq(expr, n, k, q, lower, nn0)
                    else:
                        result = QkTransform(0 * q)

                result.multiply(const)
                return result

        # Handle * exp(j*a*n+b)
        elif is_multiplied_with(expr, n, 'exp(n)', xn_fac) and abs(xn_fac[-1] / sym.exp(args[0].coeff(n, 0))) == 1:
            expr /= xn_fac[-1]
            expr = sym.simplify(expr)
            ref = xn_fac[-1].args
            aa = sym.expand(ref[0]).coeff(n, 1) / sym.I
            bb = sym.expand(ref[0]).coeff(n, 0)
            # Find transform
            result = self.termXq(expr, n, k, q, lower, upper)
            # Check frequency
            if aa.is_constant() and abs(aa) > pi:
                warn("Frequency may be out of range")

            result.subs(q, q * sym.exp(sym.I * aa))
            # Check special case and shift accordingly
            k0 = aa * self.N / 2 / pi
            if k0.is_integer and result.has_special:
                result.shift_k(k0)
            else:
                result.rm_cases()
            result.multiply(const * sym.exp(bb))
            return result

        # Handle * sin(b*n+c)
        elif is_multiplied_with(expr, n, 'sin(n)', xn_fac):
            expr /= xn_fac[-1]
            ref = xn_fac[-1].args
            bb = ref[0].coeff(n, 1)
            cc = ref[0].coeff(n, 0)
            # Check frequency
            if bb.is_constant() and abs(bb) > pi:
                warn("Frequency may be out of range")

            result = self.termXq(expr, n, k, q, lower, upper)
            # Make copies for the transformation:
            # X(q) = (exp(j*bb)*Xq (q*exp(j*bb)) - exp(-j*bb)*Xq(q*exp(-j*bb))) / 2 / j
            rq1 = deepcopy(result)
            rq2 = deepcopy(result)
            # Make general shift
            rq1.subs(q, q * sym.exp(sym.I * bb))
            rq2.subs(q, q * sym.exp(-sym.I * bb))
            # Check special case shift s
            k0 = bb * self.N / 2 / pi
            if k0.is_integer and result.has_special:
                rq1.shift_k(k0)
                rq2.shift_k(-k0)
            else:
                rq1.rm_cases()
                rq2.rm_cases()

            rq1.multiply(const * sym.exp(sym.I * cc) / 2 / sym.I)
            rq2.multiply(-const * sym.exp(-sym.I * cc) / 2 / sym.I)

            # Add both parts
            rq1.add(rq2)
            return rq1

        # Handle * cos(b*n+c)
        elif is_multiplied_with(expr, n, 'cos(n)', xn_fac):
            expr /= xn_fac[-1]
            ref = xn_fac[-1].args
            bb = ref[0].coeff(n, 1)
            cc = ref[0].coeff(n, 0)
            # Check frequency
            if bb.is_constant() and abs(bb) > pi:
                warn("Frequency may be out of range")

            result = self.termXq(expr, n, k, q, lower, upper)
            # Make copies for the transformation:
            # X(q) = (exp(j*bb)*Xq (q*exp(j*bb)) + exp(-j*bb)*Xq(q*exp(-j*bb))) / 2
            rq1 = deepcopy(result)
            rq2 = deepcopy(result)
            # Make general shifts
            rq1.subs(q, q * sym.exp(sym.I * bb))
            rq2.subs(q, q * sym.exp(-sym.I * bb))

            # Check special case shifts
            k0 = bb * self.N / 2 / pi
            if k0.is_integer and result.has_special:
                rq1.shift_k(k0)
                rq2.shift_k(-k0)
            else:
                rq1.rm_cases()
                rq2.rm_cases()
            rq1.multiply(const * sym.exp(sym.I * cc) / 2)
            rq2.multiply(const * sym.exp(-sym.I * cc) / 2)

            # Add both parts
            rq1.add(rq2)
            return rq1

        # Handle * exp(bb*n+cc)
        elif is_multiplied_with(expr, n, 'exp(n)', xn_fac):
            expr /= xn_fac[-1]
            expr = sym.simplify(expr)
            ref = xn_fac[-1].args
            bb = ref[0].coeff(n, 1)
            cc = ref[0].coeff(n, 0)
            result = self.termXq(expr, n, k, q, lower, upper)
            result.subs(q, q * sym.exp(bb))
            # No special cases remain
            result.rm_cases()
            result.multiply(const * sym.exp(cc))
            return result

        # Handle * a**n
        elif is_multiplied_with(expr, n, 'a**n', xn_fac):
            expr /= xn_fac[-1]
            expr = sym.simplify(expr)
            ref = xn_fac[-1].args
            lam = ref[0]
            bb = ref[1].coeff(n, 1)
            cc = ref[1].coeff(n, 0)
            result = self.termXq(expr, n, k, q, lower, upper)
            result.subs(q, q * lam**bb)
            # No special cases remain
            result.rm_cases()
            result.multiply(const * lam**cc)
            return result

        # Handle * n
        elif is_multiplied_with(expr, n, 'n', xn_fac):
            expr = expr / xn_fac[-1]
            result = self.termXq(expr, n, k, q, lower, upper)
            result.Xq = q * sym.diff(result.Xq, q)
            if result.has_special:
                raise ValueError(
                    "No sym.diff possible for discrete cases, refine handles")
            else:
                return result

        # No case matches so return None as this can be used to check if
        # a case was found and if a simplify is useful for the general
        # summation term simplify runs too long
        return None

    def termXk(self, expr, n, k):
        """Make transform  X = sum_0^N-1  x[n] * exp(-2*pi*j*k*n/N)
           BUT result can not be written as function of exp(-2*pi*j*k/N)
           This part will be mainly important and used with the inverse DFT

           In general the terms implemented have the form:

              exp(j*2*pi/N*n*mu)/(1-a*exp(j*2*pi/N*n))**p      with mu integer

              or exp(-j*2*pi/N*k*mu)/(1-a*exp(-j*2*pi/N*k))**p
        """

        const, expr = factor_const(expr, n)

        qq = sym.Symbol('qq')
        p0 = sym.Wild('p0', properties=[
                      lambda r: r.is_integer and r.is_nonnegative])

        if not self.is_inverse:
            sg = 1
        else:
            sg = -1

        # Replace exp(j*2*pi/N*n) by qq (forward) on calling termXk
        # for the inverse, n is k and k is n
        expr_qq = expr.replace(
            sym.exp(sg*sym.I * 2 * pi * n / self.N * p0), qq**p0)
        # kn is (always) the output variable
        kn = k

        # TODO make polynom in qq, 1/qq terms may occur?

        # First case, Denom is never zero (i.e. for no integer n)
        if not expr_qq.has(n) and expr_qq.is_rational_function(qq):
            # Find numerator and denominator
            Num = sym.numer(expr_qq)
            Denom = sym.denom(expr_qq)

            # Polynomial division  d0+d1*q+..+dN*q**N + rem/Denom
            pol, rem = sym.div(Num, Denom)
            poly_coeff = sym.Poly(pol, qq).all_coeffs()[-1::-1]

            # First part c0+c1*q+....
            result_0 = 0
            # Polynom
            for i, xx in enumerate(poly_coeff):
                result_0 += self.N * xx * UnitImpulse(kn-i) / 2

            # Second part
            #   Denominator is linear D=1-a*qq
            if Denom.as_poly(qq).is_linear:
                a0 = sym.expand(Denom).coeff(qq, 1)
                c0 = sym.expand(Denom).coeff(qq, 0)
                # Remainder
                aa = -a0 / c0
                result = rem * self.N * aa**k / (1 - aa**self.N) / c0 / 2
                return 2 * const * (result_0 + result)

            # Denominator is D=(1-a*qq)**p, p>=2
            elif Denom.is_Pow and Denom.args[0].as_poly(qq).is_linear and Denom.args[1].is_integer and Denom.args[1] >= 2:
                a0 = sym.expand(Denom.args[0]).coeff(qq, 1)
                c0 = sym.expand(Denom.args[0]).coeff(qq, 0)
                aa = -a0 / c0

                aN = aa**self.N
                NN = self.N
                pp = Denom.args[1]
                # The polynomials in k of order pp-1 of the solution (see manuscript)
                if pp == 2:
                    Q_p_mu = [(NN-1)*aN + 1 + kn*(1 - aN), NN*aN + kn*(1 - aN)]
                elif pp == 3:
                    a2N = aN**2
                    N2 = NN**2
                    Q_p_mu = [N2*a2N + N2*aN - 3*NN*a2N + 3*NN*aN + 2*a2N - 4*aN + 2 + kn * (-2*NN*a2N + 2*NN*aN + 3*a2N - 6*aN + 3) + kn**2*(aN - 1)**2,
                              N2*a2N + N2*aN - NN*a2N + NN*aN + kn *
                              (-2*NN*a2N + 2*NN*aN + a2N -
                               2*aN + 1) + kn**2*(aN - 1)**2,
                              N2*a2N + N2*aN + NN*a2N - NN*aN + kn*(-2*NN*a2N + 2*NN*aN - a2N + 2*aN - 1) + kn**2*(aN-1)**2]
                elif pp == 4:
                    a2N = aN**2
                    a3N = aN**3
                    N2 = NN**2
                    N3 = NN**3
                    Q_p_mu = [N3*a3N + 4*N3*a2N + N3*aN - 6*N2*a3N + 6*N2*aN + 11*NN*a3N - 22*NN*a2N + 11*NN*aN - 6*a3N + 18*a2N - 18*aN + 6 +
                              kn**3*(-a3N + 3*a2N - 3*aN + 1) +
                              kn**2*(3*NN*a3N - 6*NN*a2N + 3*NN*aN - 6*a3N + 18*a2N - 18*aN + 6) +
                              kn*(-3*N2*a3N + 3*N2*aN + 12*NN*a3N - 24*NN *
                                  a2N + 12*NN*aN - 11*a3N + 33*a2N - 33*aN + 11),
                              N3*a3N + 4*N3*a2N + N3*aN - 3*N2*a3N + 3*N2*aN + 2*NN*a3N - 4*NN*a2N + 2*NN*aN +
                              k**3*(-a3N + 3*a2N - 3*aN + 1) +
                              k**2*(3*NN*a3N - 6*NN*a2N + 3*NN*aN - 3*a3N + 9*a2N - 9*aN + 3) +
                              k*(-3*N2*a3N + 3*N2*aN + 6*NN*a3N - 12*NN *
                                 a2N + 6*NN*aN - 2*a3N + 6*a2N - 6*aN + 2),
                              N3*a3N + 4*N3*a2N + N3*aN - NN*a3N + 2*NN*a2N - NN*aN +
                              k**3*(-a3N + 3*a2N - 3*aN + 1) +
                              k**2*(3*NN*a3N - 6*NN*a2N + 3*NN*aN) +
                              k*(-3*N2*a3N + 3*N2*aN + a3N - 3*a2N + 3*aN - 1),
                              N3*a3N + 4*N3*a2N + N3*aN + 3*N2*a3N - 3*N2*aN + 2*NN*a3N - 4*NN*a2N + 2*NN*aN +
                              k**3*(-a3N + 3*a2N - 3*aN + 1) +
                              k**2*(3*NN*a3N - 6*NN*a2N + 3*NN*aN + 3*a3N - 9*a2N + 9*aN - 3) +
                              k*(-3*N2*a3N + 3*N2*aN - 6*NN*a3N + 12*NN*a2N - 6*NN*aN - 2*a3N + 6*a2N - 6*aN + 2)]
                elif pp == 5:
                    a2N = aN**2
                    a3N = aN**3
                    a4N = aN**4
                    N2 = NN**2
                    N3 = NN**3
                    N4 = NN**4
                    Q_p_mu = [N4*a4N + 11*N4*a3N + 11*N4*a2N + N4*aN - 10*N3*a4N - 30*N3*a3N + 30*N3*a2N + 10*N3*aN + 35*N2*a4N - 35*N2*a3N - 35*N2*a2N + 35*N2*aN - 50*NN*a4N + 150*NN*a3N - 150*NN*a2N + 50*NN*aN + 24*a4N - 96*a3N + 144*a2N - 96*aN + 24 +
                              kn**4*(a4N - 4*a3N + 6*a2N - 4*aN + 1) +
                              kn**3*(-4*NN*a4N + 12*NN*a3N - 12*NN*a2N + 4*NN*aN + 10*a4N - 40*a3N + 60*a2N - 40*aN + 10) +
                              kn**2*(6*N2*a4N - 6*N2*a3N - 6*N2*a2N + 6*N2*aN - 30*NN*a4N + 90*NN*a3N - 90*NN*a2N + 30*NN*aN + 35*a4N - 140*a3N + 210*a2N - 140*aN + 35) +
                              kn*(-4*N3*a4N - 12*N3*a3N + 12*N3*a2N + 4*N3*aN + 30*N2*a4N - 30*N2*a3N - 30*N2*a2N + 30*N2 *
                                  aN - 70*NN*a4N + 210*NN*a3N - 210*NN*a2N + 70*NN*aN + 50*a4N - 200*a3N + 300*a2N - 200*aN + 50),
                              N4*a4N + 11*N4*a3N + 11*N4*a2N + N4*aN - 6*N3*a4N - 18*N3*a3N + 18*N3*a2N + 6*N3*aN + 11*N2*a4N - 11*N2*a3N - 11*N2*a2N + 11*N2*aN - 6*NN*a4N + 18*NN*a3N - 18*NN*a2N + 6*NN*aN + k**4*(a4N - 4*a3N + 6*a2N - 4*aN + 1) +
                              kn**3*(-4*NN*a4N + 12*NN*a3N - 12*NN*a2N + 4*NN*aN + 6*a4N - 24*a3N + 36*a2N - 24*aN + 6) +
                              kn**2*(6*N2*a4N - 6*N2*a3N - 6*N2*a2N + 6*N2*aN - 18*NN*a4N + 54*NN*a3N - 54*NN*a2N + 18*NN*aN + 11*a4N - 44*a3N + 66*a2N - 44*aN + 11) +
                              kn*(-4*N3*a4N - 12*N3*a3N + 12*N3*a2N + 4*N3*aN + 18*N2*a4N - 18*N2*a3N - 18*N2*a2N + 18 *
                                  N2*aN - 22*NN*a4N + 66*NN*a3N - 66*NN*a2N + 22*NN*aN + 6*a4N - 24*a3N + 36*a2N - 24*aN + 6),
                              N4*a4N + 11*N4*a3N + 11*N4*a2N + N4*aN - 2*N3*a4N - 6*N3*a3N + 6*N3*a2N + 2*N3*aN - N2*a4N + N2*a3N + N2*a2N - N2*aN + 2*NN*a4N - 6*NN*a3N + 6*NN*a2N - 2*NN*aN +
                              kn**4*(a4N - 4*a3N + 6*a2N - 4*aN + 1) +
                              kn**3*(-4*NN*a4N + 12*NN*a3N - 12*NN*a2N + 4*NN*aN + 2*a4N - 8*a3N + 12*a2N - 8*aN + 2) +
                              kn**2*(6*N2*a4N - 6*N2*a3N - 6*N2*a2N + 6*N2*aN - 6*NN*a4N + 18*NN*a3N - 18*NN*a2N + 6*NN*aN - a4N + 4*a3N - 6*a2N + 4*aN - 1) +
                              kn*(-4*N3*a4N - 12*N3*a3N + 12*N3*a2N + 4*N3*aN + 6*N2*a4N - 6*N2*a3N - 6*N2*a2N + 6 *
                                  N2*aN + 2*NN*a4N - 6*NN*a3N + 6*NN*a2N - 2*NN*aN - 2*a4N + 8*a3N - 12*a2N + 8*aN - 2),
                              N4*a4N + 11*N4*a3N + 11*N4*a2N + N4*aN + 2*N3*a4N + 6*N3*a3N - 6*N3*a2N - 2*N3*aN - N2*a4N + N2*a3N + N2*a2N - N2*aN - 2*NN*a4N + 6*NN*a3N - 6*NN*a2N + 2*NN*aN +
                              kn**4*(a4N - 4*a3N + 6*a2N - 4*aN + 1) + k**3*(-4*NN*a4N + 12*NN*a3N - 12*NN*a2N + 4*NN*aN - 2*a4N + 8*a3N - 12*a2N + 8*aN - 2) +
                              kn**2*(6*N2*a4N - 6*N2*a3N - 6*N2*a2N + 6*N2*aN + 6*NN*a4N - 18*NN*a3N + 18*NN*a2N - 6*NN*aN - a4N + 4*a3N - 6*a2N + 4*aN - 1) +
                              kn*(-4*N3*a4N - 12*N3*a3N + 12*N3*a2N + 4*N3*aN - 6*N2*a4N + 6*N2*a3N + 6*N2*a2N - 6 *
                                  N2*aN + 2*NN*a4N - 6*NN*a3N + 6*NN*a2N - 2*NN*aN + 2*a4N - 8*a3N + 12*a2N - 8*aN + 2),
                              N4*a4N + 11*N4*a3N + 11*N4*a2N + N4*aN + 6*N3*a4N + 18*N3*a3N - 18*N3*a2N - 6*N3*aN + 11*N2*a4N - 11*N2*a3N - 11*N2*a2N + 11*N2*aN + 6*NN*a4N - 18*NN*a3N + 18*NN*a2N - 6*NN*aN +
                              kn**4*(a4N - 4*a3N + 6*a2N - 4*aN + 1) + k**3*(-4*NN*a4N + 12*NN*a3N - 12*NN*a2N + 4*NN*aN - 6*a4N + 24*a3N - 36*a2N + 24*aN - 6) +
                              kn**2*(6*N2*a4N - 6*N2*a3N - 6*N2*a2N + 6*N2*aN + 18*NN*a4N - 54*NN*a3N + 54*NN*a2N - 18*NN*aN + 11*a4N - 44*a3N + 66*a2N - 44*aN + 11) +
                              kn*(-4*N3*a4N - 12*N3*a3N + 12*N3*a2N + 4*N3*aN - 18*N2*a4N + 18*N2*a3N + 18*N2*a2N - 18*N2*aN - 22*NN*a4N + 66*NN*a3N - 66*NN*a2N + 22*NN*aN - 6*a4N + 24*a3N - 36*a2N + 24*aN - 6)]

                else:
                    # TODO  make faster
                    # we increase performance for pp >= 6 by saving the table in an extra file and read the polynoms and subs
                    # or make make_polynom_table1 faster by substituting a**(r*N) by arN and so on
                    Q_p_mu = make_polynom_table1(kn, pp, aa, NN)

                # case   d0/(1-a*qq)**p   + d1*qq/(1-a*qq)**p + .... + dm *qq**(p-1)/(1-a*qq)**p
                num_coeff = sym.Poly(rem, qq).all_coeffs()[-1::-1]
                if len(num_coeff) > len(Q_p_mu):
                    raise ValueError(
                        " Error : polynomial division may be wrong! ")
                # sum up
                result = 0
                for i in range(len(num_coeff)):
                    result += num_coeff[i] * Q_p_mu[i] / aa**i
                result *= NN / \
                    sym.factorial(pp - 1) * aa**kn / \
                    (1 - aN) ** pp / c0**pp / 2
                return 2 * const * (result_0 + result)

            else:
                return None

        # Second case, denominators has roots for integer n,
        # so term must contain a factor (1-delta(k-k0)) to indicate that k!=k0
        # better the expression is defined with sym.Piecewise sym.Piecewise((expr, sym.Ne(k, k0)))
        # search for a factor (1-delta(n-n0)) or for piecewise.
        this_case = False
        # Check for piecewise
        if expr.is_Piecewise:
            match = list(expr.args[0][1].find(sym.Ne(n, p0)))
            n0 = match[0].args[1]
            div_expr = expr_qq.args[0][0]
            this_case = True
        # Check for (1-delta) factor
        else:
            match = list(expr.find(1 - UnitImpulse(n - p0)))
            if len(match) == 1:
                n0 = -((-match[0] + 1).args[0] - n)
                div_expr = expr_qq / match[0]
                this_case = True
        if this_case:
            if not div_expr.has(n) and div_expr.is_rational_function(qq):
                # Find numerator and denominator
                Num = sym.numer(div_expr)
                Denom = sym.denom(div_expr)

                # Polynomial division  d0+d1*q+..+dN*q**N + rem/N
                pol, rem = sym.div(Num, Denom)
                poly_coeff = sym.Poly(pol, qq).all_coeffs()[-1::-1]

                # First part c0+c1*q+....
                result_0 = 0
                # Polynom
                for i, xx in enumerate(poly_coeff):
                    # as n!=n0 this n has to be excluded
                    if i != n0:
                        result_0 += xx * UnitImpulse(kn - i) * self.N

                # Denominator is linear D=1-a*qq
                if Denom.as_poly(qq).is_linear:
                    a0 = sym.expand(Denom).coeff(qq, 1)
                    c0 = sym.expand(Denom).coeff(qq, 0)
                    # remainder
                    aa = -a0 / c0
                    if not aa == sym.exp(-sg * j * 2 * pi / self.N * n0):
                        print("Rewrite expression as ", expr /
                              match[0] - expr.subs(n, n0))
                        return None
                    # Result
                    result = ((self.N - 1) / 2 - kn) * \
                        sym.exp(-sg * j * 2 * pi * k * n0 / self.N) / c0
                    return const * (result_0 + result)

                # Denominator is D=(1-a*qq)**p, p>=2
                elif Denom.is_Pow and Denom.args[0].as_poly(qq).is_linear and Denom.args[1].is_integer and Denom.args[1] >= 2:
                    a0 = sym.expand(Denom.args[0]).coeff(qq, 1)
                    c0 = sym.expand(Denom.args[0]).coeff(qq, 0)
                    aa = -a0 / c0
                    if not aa == sym.exp(-sg * j * 2 * pi / self.N * n0):
                        print("Rewrite expression as ", expr /
                              match[0] - expr.subs(n, n0))
                        return None

                    NN = self.N
                    pp = Denom.args[1]
                    # The polynomials in k of order pp of the solution (see manuscript)
                    if pp == 2:
                        T_p_mu = [(-NN**2 + 6*NN - 6*kn**2 + 6*kn*(NN - 2) - 5) / 12,
                                  (-NN**2 + 6*NN*kn - 6*kn**2 + 1) / 12]
                    elif pp == 3:
                        T_p_mu = [(-3*NN**2 + 12*NN - 4*kn**3 + 6*kn**2*(NN - 3) + kn*(-2*NN**2 + 18*NN - 24) - 9) / 24,
                                  (-NN**2 - 4*kn**3 + 6*kn**2*(NN - 1) +
                                   kn*(-2*NN**2 + 6*NN) + 1) / 24,
                                  (NN**2 - 4*kn**3 + 6*kn**2*(NN + 1) + kn*(-2*NN**2 - 6*NN) - 1) / 24]
                    elif pp == 4:
                        T_p_mu = [(NN**4 - 110*NN**2 + 360*NN - 30*kn**4 + kn**3*(60*NN - 240) + kn**2*(-30*NN**2 + 360*NN - 660) + kn*(-120*NN**2 + 660*NN - 720) - 251) / 720,
                                  (NN**4 - 20*NN**2 - 30*kn**4 + kn**3*(60*NN - 120) + kn**2 *
                                   (-30*NN**2 + 180*NN - 120) + kn*(-60*NN**2 + 120*NN) + 19) / 720,
                                  (NN**4 + 10*NN**2 + 60*NN*kn**3 - 60*NN*kn -
                                   30*kn**4 + kn**2*(60 - 30*NN**2) - 11) / 720,
                                  (NN**4 - 20*NN**2 - 30*kn**4 + kn**3*(60*NN + 120) + kn**2*(-30*NN**2 - 180*NN - 120) + kn*(60*NN**2 + 120*NN) + 19) / 720]
                    elif pp == 5:
                        T_p_mu = [(5*NN**4 - 250*NN**2 + 720*NN - 12*kn**5 + kn**4*(30*NN - 150) + kn**3*(-20*NN**2 + 300*NN - 700) + kn**2*(-150*NN**2 + 1050*NN - 1500) + kn*(2*NN**4 - 350*NN**2 + 1500*NN - 1440) - 475) / 1440,
                                  (3*NN**4 - 30*NN**2 - 12*kn**5 + kn**4*(30*NN - 90) + kn**3*(-20*NN**2 + 180*NN - 220) +
                                   kn**2*(-90*NN**2 + 330*NN - 180) + kn*(2*NN**4 - 110*NN**2 + 180*NN) + 27) / 1440,
                                  (NN**4 + 10*NN**2 - 12*kn**5 + kn**4*(30*NN - 30) + kn**3*(-20*NN**2 + 60*NN + 20) +
                                   kn**2*(-30*NN**2 - 30*NN + 60) + kn*(2*NN**4 + 10*NN**2 - 60*NN) - 11) / 1440,
                                  (-NN**4 - 10*NN**2 - 12*kn**5 + kn**4*(30*NN + 30) + kn**3*(-20*NN**2 - 60*NN + 20) + kn**2*(
                                      30*NN**2 - 30*NN - 60) + kn*(2*NN**4 + 10*NN**2 + 60*NN) + 11) / 1440,
                                  (-3*NN**4 + 30*NN**2 - 12*kn**5 + kn**4*(30*NN + 90) + kn**3*(-20*NN**2 - 180*NN - 220) + kn**2*(90*NN**2 + 330*NN + 180) + kn*(2*NN**4 - 110*NN**2 - 180*NN) - 27) / 1440]
                    else:
                        T_p_mu = make_polynom_table2(kn, pp, varN=NN)

                    num_coeff = sym.Poly(rem, qq).all_coeffs()[-1::-1]
                    if len(num_coeff) > len(T_p_mu):
                        raise ValueError(
                            "Error: polynomial division may be wrong!")
                    # Sum up
                    result = 0
                    for i in range(len(num_coeff)):
                        result += num_coeff[i] * T_p_mu[i] * \
                            sym.exp(j * 2 * pi * n0 / self.N * i * sg)
                    result *= sym.exp(-j * 2 * pi * n0 /
                                      self.N * k * sg) / c0**pp

                    return const * (result_0 + result)
        return None

    def term(self, expr, n, k, **kwargs):

        const, expr = factor_const(expr, n)
        if self.is_inverse:
            const /= self.N

        if expr.has(AppliedUndef):
            # Handle v(n), v(n) * y(n), 3 * v(n) / n etc.
            result = const * self.function(expr, n, k)
            return result

        # Transforms of type x[n] o--o X(q) with q=exp(-j*2*pi/N*k)
        # (mainly used with the forward transform, but works also for
        # backward transforms)

        # Call transform, use inversion for first case (change sign
        # later back for second case)
        k = -k if self.is_inverse else k

        q = sym.Symbol('q')
        res = self.termXq(expr, n, k, q, 0, self.N - 1)
        # Find handles
        if not res is None:
            # Simplify q**N terms
            res.simp_qN(q, self.N)
            # Make final transform by putting all cases and terms together
            res.make_transform(self.N, self.is_inverse, q,
                               sym.exp(-sym.I * 2 * pi / self.N * k), k,
                               kwargs.get('piecewise', None))
            result = res.Xk

            # TODO smart simplification of result since simplify takes
            # too long or makes sometimes weird expressions might be
            # better in make_transform

            # result = sym.simplify(result)   #only useful if handles found
            return const * result

        # change back, as next handles do not need  replacement k -> -k
        if self.is_inverse:
            k = -k

        # Transforms of type x[n] = rational function(qq)  with qq=exp(j*2*pi/N*n)
        # (mainly used with the backward transform, but works also for forward transform)
        result = self.termXk(expr, n, k)
        if not result is None:
            return const * result

        # No special handle identified, general expression
        k = -k if self.is_inverse else k
        if not self.N.is_symbol:
            result = sym.summation(
                expr * sym.exp(-j * 2 * pi * n * k / self.N), (n, 0, self.N - 1))
        else:
            result = sym.Sum(expr * sym.exp(-j * 2 * pi *
                             n * k / self.N), (n, 0, self.N - 1))
        return const * result


dft_transformer = DFTTransformer()


def discrete_fourier_transform(expr, n, k, N=None, evaluate=True,
                               **kwargs):
    """Compute bilateral discrete Fourier transform of expr.

    Undefined functions such as x(n) are converted to X(k)
    """

    return dft_transformer.transform(expr, n, k, evaluate=evaluate, N=N,
                                     **kwargs)


def DFT(expr, n, k, N=None, evaluate=True, **kwargs):
    """Compute bilateral discrete Fourier transform of expr.

    Undefined functions such as x(n) are converted to X(k)
    """

    return dft_transformer.transform(expr, n, k, evaluate=evaluate, N=N,
                                     **kwargs)


def DFTmatrix(N):
    """Return DFT matrix of size `N` x `N`."""

    from .functions import exp
    from .sym import j, pi

    w = exp(-j * 2 * pi / N)

    a = Matrix.zeros(N)

    for row in range(N):
        for col in range(N):
            a[row, col] = w ** (row * col)
    return a



from lcapy import *
import unittest


class LcapyTester(unittest.TestCase):

    """Unit tests for lcapy

    """

    def test_schematic1(self):

        a = R(1) + C(2)
        a.draw('tmp.tex')

        match = r"""\documentclass[a4paper]{standalone}
\usepackage{amsmath}
\usepackage{circuitikz}
\usetikzlibrary{fit, shapes, arrows, patterns, decorations.text, decorations.markings}

\begin{document}
\begin{tikzpicture}[scale=1.00, transform shape, /tikz/circuitikz/bipoles/length=1.50cm, american currents, american voltages, voltage dir=RP]
  \coordinate (1) at (0,0);
  \coordinate (2) at (2,0);
  \coordinate (3) at (2.5,0);
  \coordinate (0) at (4.5,0);
  \draw (1) to [R, l_=$1\,\mbox{$\Omega$}$, n=R1] (2);
  \draw[-] (2) to (3);
  \draw (3) to [C, l_=$2\,\mbox{F}$, n=C1] (0);
  \draw (1) node[ocirc] {};
  \draw (0) node[ocirc] {};
\end{tikzpicture}
\end{document}"""

        content = open('tmp.tex').read()
        if content != match:
            raise ValueError('Schematic mismatch')

    def test_schematic2(self):

        a = Circuit("""
R1 1 2 1; right
R2 2 3 2; right
C 1 3 3; right""")
        a.draw('tmp.tex')

        match = r"""\documentclass[a4paper]{standalone}
\usepackage{amsmath}
\usepackage{circuitikz}
\usetikzlibrary{fit, shapes, arrows, patterns, decorations.text, decorations.markings}

\begin{document}
\begin{tikzpicture}[scale=1.00, transform shape, /tikz/circuitikz/bipoles/length=1.50cm, american currents, american voltages, voltage dir=RP]
  \coordinate (1) at (0,0);
  \coordinate (2) at (2,0);
  \coordinate (3) at (4,0);
  \draw (1) to [R, l_={$R_{1}$}{=$1\,\mbox{$\Omega$}$}, n=R1] (2);
  \draw (2) to [R, l_={$R_{2}$}{=$2\,\mbox{$\Omega$}$}, n=R2] (3);
  \draw (1) to [C, l_={$C$}{=$3\,\mbox{F}$}, n=C] (3);
  \draw (1) node[circ] {};
  \draw (2) node[circ] {};
  \draw (3) node[circ] {};
  \draw[anchor=south east] (1) node {1};
  \draw[anchor=south east] (2) node {2};
  \draw[anchor=south east] (3) node {3};
\end{tikzpicture}
\end{document}"""

        content = open('tmp.tex').read()
        if content != match:
            raise ValueError('Schematic mismatch')

    def test_network_node_positions1(self):

        a = R(1) + C(2)
        sch = a.sch()
        sch._positions_calculate()
        nodes = sch.nodes

        self.assertEqual(nodes['1'].pos.x, 0, '1 x')
        self.assertEqual(nodes['2'].pos.x, 2, '2 x')
        self.assertEqual(nodes['3'].pos.x, 2.5, '3 x')
        self.assertEqual(nodes['0'].pos.x, 4.5, '0 x')

    def test_network_node_positions2(self):

        a = (R(1) + C(2)) | L(3)
        sch = a.sch()
        sch._positions_calculate()
        nodes = sch.nodes

        self.assertEqual(nodes['1'].pos.x, 0, '1 x')
        self.assertEqual(nodes['2'].pos.x, 0.5, '2 x')
        self.assertEqual(nodes['3'].pos.x, 5, '3 x')
        self.assertEqual(nodes['4'].pos.x, 0.5, '4 x')
        self.assertEqual(nodes['5'].pos.x, 5, '5 x')
        self.assertEqual(nodes['6'].pos.x, 2.5, '6 x')
        self.assertEqual(nodes['7'].pos.x, 3, '7 x')
        self.assertEqual(nodes['8'].pos.x, 0.5, '8 x')
        self.assertEqual(nodes['9'].pos.x, 5, '9 x')
        self.assertEqual(nodes['0'].pos.x, 5.5, '0 x')

    def test_network_danglin3(self):

        a = Circuit("""
R1 1 2; right
R2 2 3; right
R3 2 5; down
R4 4 5; right=2
R5 5 6; right
R6 5 7; down
R7 7 8; right=2, fixed""")

        sch = a.sch
        sch._positions_calculate()
        nodes = sch.nodes

        self.assertEqual(nodes['1'].pos.x, 2, '1 x')
        self.assertEqual(nodes['2'].pos.x, 4, '2 x')
        self.assertEqual(nodes['3'].pos.x, 6, '3 x')
        self.assertEqual(nodes['4'].pos.x, 0, '4 x')
        self.assertEqual(nodes['5'].pos.x, 4, '5 x')
        self.assertEqual(nodes['6'].pos.x, 6, '6 x')
        self.assertEqual(nodes['7'].pos.x, 4, '7 x')
        self.assertEqual(nodes['8'].pos.x, 8, '8 x')



from .domains import domains
from .quantities import quantities
from .exprclasses import exprclasses


def exprmap(quantity, domain):

    domaincls = domains[domain]

    if quantity == 'undefined':
        return exprclasses[domain][quantity]

    quantitycls = quantities[quantity]

    if (domaincls.is_constant_domain or domaincls.is_constant_time_domain):
        if quantitycls.is_ratio:
            domain = 'constant frequency response'

    if (domaincls.is_constant_domain or domaincls.is_constant_frequency_response_domain):
        if not quantitycls.is_ratio:
            domain = 'constant time'

    return exprclasses[domain][quantity]



"""This module provides the SuperpositionDomainExpression class.  It
is the base class for Current and Voltage.  It represents voltages and
currents as a superposition in different transform domains.

Copyright 2019--2023 Michael Hayes, UCECE

"""

from __future__ import division
import sys
import sympy as sym
from .expr import Expr, ExprDict, expr
from .sym import tsym, omega0sym, symbols_find, is_sympy
from .acdc import is_ac
from .state import state
from .domains import SuperpositionDomain
from .printing import pprint, pretty, latex
from .classmap import domain_kind_quantity_to_class, expr_to_domain_kind
from .classmap import domain_quantity_to_class


class Superposition(SuperpositionDomain, ExprDict):
    """This class represents a superposition of different signal types,
    DC, AC, transient, and noise.

    The time-domain representation is returned with the time method,
    V.time(), or with the notation V(t).  This does not include the
    noise component.

    The Laplace representation is returned with the laplace method,
    V.laplace() or with the notation V(s).  This does not include the
    noise component.

    Noise components with different noise identifiers are stored
    separately, keyed by the noise identifier.  They are ignored by
    the laplace() and time() methods.

    The total noise can be accessed with the .n attribute.  This sums
    each of the noise components in quadrature since they are
    independent.

    For example, consider V = Voltage('cos(3 * t) + exp(-4 * t) + 5')

    str(V(t)) gives 'cos(3*t) + 5 + exp(-4*t)'

    str(V(s)) gives 's/(9*(s**2/9 + 1)) + 1/(4*(s/4 + 1)) + 5/s'

    V.dc gives 5

    V.ac gives {3: 1}

    V.s gives 1/(s + 4)
    """

    # TODO: rework this class to better show 0 result.

    # Where possible this class represents a signal in the time-domain.
    # It can decompose a signal into AC, DC, and transient components.
    # The 't' key is the transient component viewed in the time domain.
    # The 's' key is the transient component viewed in the Laplace domain.

    def __init__(self, *args, **kwargs):
        super(Superposition, self).__init__()

        for arg in args:

            if not isinstance(arg, (Superposition)):
                arg = expr(arg).as_quantity(self.quantity)

            self.add(arg)

    def _representation(self):
        if not any(self):
            return 0

        if len(self) == 1:
            decomp = self
        else:
            decomp = self.decompose()

        # If have a single element, show value.
        if len(decomp) == 1:
            key = list(decomp)[0]
            if key in ('dc', 't'):
                return decomp[key]
            if key in ('s', 'f', 'omega'):
                expr = decomp[key]
                if expr.has(expr.var):
                    return decomp[key]
                # Have something like s * 0 + 1 so don't display as 1
                # since not a dc value.

        # Have a superposition.
        # It is probably less confusing to a user to display
        # using a decomposition in the transform domains.
        # We could present the result in the time-domain but this
        # hides the underlying way the signal is analysed.
        return dict(decomp)

    def latex(self, **kwargs):
        """Latex"""
        return latex(self._representation(), **kwargs)

    def _repr_pretty_(self, p, cycle):
        """This is used by jupyter notebooks to display an expression using
        unicode."""

        p.text(pretty(self._representation()))

    def _repr_latex_(self):
        """This is used by jupyter notebooks to display an expression using
        LaTeX markup.  However, this requires mathjax.  If this method
        is not defined, jupyter falls back on _repr__pretty_ which
        outputs unicode."""

        s = latex(self._representation(), mode='plain')
        return "$$%s$$" % s

    def _repr_png_(self):

        return None

    def _repr_svg_(self):

        return None

    def pprint(self):
        """Pretty print"""

        return pprint(self._representation())

    def __getitem__(self, key):
        # This allows a[omega] to work if omega used as key
        # instead of 'omega'.
        if isinstance(key, Expr):
            key = key.expr
        if key in self:
            return super(Superposition, self).__getitem__(key)
        decomp = self.decompose()
        if key == 0 and 'dc' in decomp:
            key = 'dc'
        if key not in decomp:
            raise IndexError('Unknown signal component %s' % key)

        return decomp.__getitem__(key)

    def decompose_to_domain(self, expr, kind):

        cls = domain_kind_quantity_to_class(kind, self.quantity)
        if cls.is_phasor_domain:
            return cls(expr, omega=kind)
        return cls(expr)

    @property
    def symbols(self):
        """Return dictionary of symbols in the expression keyed by name."""

        syms = {}
        for expr1 in self.values():
            syms.update(expr1.symbols)
        return syms

    @property
    def units(self):
        """Return the units of the expression.  However, since this is a
        superposition, units are not defined.  Instead, it is
        necessary to transform to a specific domain.  For example,
        using `V(s)`."""

        raise AttributeError(
            """Units are not defined for a superposition; convert to a specific domain, say using `V(t)` or `V(s)`.""")

    def ac_keys(self):
        """Return list of keys for all ac components."""

        keys = []
        for key in self.decompose().keys():
            if not isinstance(key, str) or key == 'w':
                keys.append(key)
        return keys

    def noise_keys(self):
        """Return list of keys for all noise components."""

        keys = []
        for key in self.keys():
            if isinstance(key, str) and key[0] == 'n':
                keys.append(key)
        return keys

    def has(self, subexpr):
        """Test whether the sub-expression is contained.  For example,
         V.has(exp(t))
         V.has(t)

        """
        for key, expr1 in self.items():
            if expr1.has(subexpr):
                return True
        return False

    def has_symbol(self, sym):
        """Test if have symbol.  For example,
        V.has_symbol('a')
        V.has_symbol(t)

        """
        return self.has(expr(sym))

    @property
    def has_dc(self):
        """True if there is a DC component."""
        return self == 0 or 'dc' in self.decompose()

    @property
    def has_ac(self):
        """True if there is an AC component."""
        return self.ac_keys() != []

    @property
    def has_s_transient(self):
        """True if have transient component defined in the s-domain."""
        return 's' in self

    @property
    def has_t_transient(self):
        """True if have transient component defined in the time-domain."""
        return 't' in self

    @property
    def has_transient(self):
        """True if have transient component."""
        return self.has_s_transient or self.has_t_transient

    @property
    def has_noisy(self):
        """True if there is a noise component."""
        return self.noise_keys() != []

    @property
    def is_dc(self):
        """True if only has a DC component."""
        return self == 0 or (self.has_dc and list(self.decompose().keys()) == ['dc'])

    @property
    def is_ac(self):
        """True if only has AC components."""
        return self.has_ac and self.ac_keys() == list(self.keys())

    @property
    def is_phasor_domain(self):
        """True if has single AC component."""

        return self.ac and len(list(self.keys())) == 1

    @property
    def is_noisy(self):
        """True if only has noise components."""
        return self.has_noisy and self.noise_keys() == list(self.keys())

    @property
    def is_s_transient(self):
        """True if only has s-domain transient component."""
        return list(self.decompose().keys()) == ['s']

    @property
    def is_t_transient(self):
        """True if only has t-domain transient component."""
        return list(self.decompose().keys()) == ['t']

    @property
    def is_transient(self):
        """True if only has transient component(s).  Note, should
        not have both t and s components."""
        return self.is_s_transient or self.is_t_transient

    @property
    def is_causal(self):
        return (self.is_transient and self.s.is_causal) or self == 0

    @property
    def is_superposition(self):
        return len(self.keys()) > 1

    def __call__(self, arg, **assumptions):
        """
        arg determines the returned representation
          t: time-domain representation
          s: Laplace domain representation
          f: Fourier representation
          omega: Fourier representation (angular frequency)
          jomega: Laplace domain representation with s = jomega

        For example, V(t), V(f), or V(2 * t).

        If arg is a constant, the time-domain representation
        evaluated at the argument is returned, for example,
        V(0) returns the dc value.
        """

        from .transform import call
        return call(self, arg, **assumptions)

    def subs(self, *args, **kwargs):

        new = self.__class__()
        for kind, value in self.items():
            new[kind] = value.subs(*args, **kwargs)
        return new

    def initial_value(self):
        """Determine value at t = 0.
        See also pre_initial_value and post_initial_value"""

        return self.time().initial_value()

    def pre_initial_value(self):
        """Determine value at t = 0-.
        See also initial_value and post_initial_value"""

        return self.time().pre_initial_value()

    def post_initial_value(self):
        """Determine value at t = 0+.
        See also pre_initial_value and initial_value"""

        return self.time().post_initial_value()

    def final_value(self):
        """Determine value at t = oo."""

        return self.time().final_value()

    def transform(self, arg, **assumptions):
        """Transform into a different domain."""

        from .transform import transform
        return transform(self, arg, **assumptions)

    def simplify(self):

        raise ValueError(
            'You need to select a specific domain, e.g., using V(t).simplify()')

    def __add__(self, x):

        def _is_s_arg(x):
            return isinstance(x, LaplaceDomainExpression) or (isinstance(x, Superposition) and 's' in x)

        x = expr(x)
        # Perhaps check if a constant?
        if x.is_undefined:
            x = x.as_quantity(self.quantity)

        if x.quantity != self.quantity:
            raise ValueError('Incompatible quantities %s and %s' %
                             (self.quantity, x.quantity))

        if _is_s_arg(x):
            new = self.decompose()
        else:
            new = self.__class__(self)

        if isinstance(x, Superposition):
            for value in x.values():
                new.add(value)
        else:
            new.add(x)
        return new

    def __radd__(self, x):
        return self.__add__(x)

    def __sub__(self, x):
        return self + (-x)

    def __rsub__(self, x):
        return -self + x

    def __neg__(self):
        new = self.__class__()
        for kind, value in self.items():
            new[kind] = -value
        return new

    def __scale__(self, x):
        new = self.__class__()
        for kind, value in self.items():
            new[kind] = value * x
        return new

    def __eq__(self, x):
        """ Test for mathematical equality as far as possible.
        This cannot be guaranteed since it depends on simplification.
        Note, SymPy comparison is for structural equality.

        Also note, noise cannot be compared."""

        diff = (self - x).simplify().decompose()
        for kind, value in diff.items():
            if value != 0:
                return False
        return True

    def __ne__(self, x):
        """ Test for mathematical inequality as far as possible.
        This cannot be guaranteed since it depends on simplification.
        Note, SymPy comparison is for structural equality.

        Also note, noise cannot be compared."""

        diff = (self - x).simplify().decompose()
        for kind, value in diff.items():
            if value != 0:
                return True
        return False

    def _decompose_timedomain_expr(self, expr1):
        """Decompose a time domain expr into AC, DC, and s-domain
        transient components."""

        result = {}

        # Extract DC components
        dc = expr1.expr.coeff(tsym, 0)
        dc = ConstantDomainExpression(dc).as_quantity(self.quantity)
        if dc != 0:
            result['dc'] = dc
            expr1 -= dc

        if expr1 == 0:
            return result

        # Extract AC components
        ac = 0
        terms = expr1.expr.as_ordered_terms()
        for term in terms:
            if is_ac(term, tsym):
                eterm = expr(term).as_quantity(self.quantity)
                p = eterm.phasor()
                result[p.omega] = p
                ac += eterm

        expr1 -= ac
        if expr1 == 0:
            return result

        # The remaining components are considered transient
        # so convert to Laplace representation.
        sval = expr1.laplace()

        result['s'] = sval
        return result

    def decompose(self):
        """Decompose into a new representation in the transform domains."""

        if hasattr(self, '_decomposition'):
            return self._decomposition

        new = self.__class__()

        for kind, value in self.items():
            if kind == 't':
                decomp = new._decompose_timedomain_expr(value)
                for value in decomp.values():
                    new.add(value)
            else:
                new.add(value)

        self._decomposition = new
        return new

    def select(self, kind):
        """Select a component of the signal representation by kind where:
        'super' : the entire superposition
        'time' :  the time domain representation (equivalent to self.time())
        'laplace' :  the laplace domain representation (equivalent to self.laplace())
        'ivp' :  the s-domain representation (equivalent to self.laplace())
        'dc' : the DC component
        omega : the AC component with angular frequency omega
        's' : the transient component in the s-domain
        'n' : the noise component
        't' : the time-domain transient component (this may or may not
              include the DC and AC components).

        """

        if isinstance(kind, str):
            if kind == 'super':
                return self
            elif kind == 'time':
                return self.time()
            elif kind in ('ivp', 'laplace'):
                return self.laplace()

            if isinstance(kind, str) and kind[0] == 'n':
                if kind not in self:
                    return self.decompose_to_domain(0, 'n')
                return self[kind]

        obj = self
        if 't' in self and (kind == omega or 't' != kind):
            # The rationale here is that there may be
            # DC and AC components included in the 't' part.
            obj = self.decompose()

        if kind not in obj:
            return obj.decompose_to_domain(0, kind)
        return obj[kind]

    def netval(self, kind):

        def kind_keyword(kind):
            if not isinstance(kind, str):
                return 'ac'

            if kind[0] == 'n':
                return 'noise'
            elif kind in ('ivp', 'laplace'):
                return 's'
            elif kind in ('t', 'time'):
                return ''
            return kind

        val = self.select(kind)
        if (isinstance(kind, str) and kind in ('s', 'ivp') and
                (val.is_causal or val.is_dc or val.is_ac)):
            # Convert to time representation so that can re-infer
            # causality, etc.
            return val.time()

        keyword = kind_keyword(kind)

        if 'nid' in val.assumptions:
            return keyword, val, val.nid

        if keyword == 'ac':
            return keyword, val, 0, val.omega

        return keyword, val

    def _netrepr(self):
        """Format value as a string value for a netlist."""

        # FIXME: converting to time-domain is only required
        # if have a superposition of values.
        return str(self.time())

    def _kind(self, value):
        if isinstance(value, PhasorDomainExpression):
            # Use angular frequency for key.  This can be a nuisance
            # for numerical values since cannot use x.3 syntax
            # say for an angular frequency of 3.
            key = value.omega
            if isinstance(key, Expr):
                key = key.expr
            return key

        kind = expr_to_domain_kind(value)
        return kind

    def _parse(self, string):
        """Parse t or s-domain expression or symbol, interpreted in time
        domain if not containing s, omega, or f.  Return most
        appropriate transform domain.

        """

        symbols = symbols_find(string)
        if 's' in symbols:
            return self.add(LaplaceDomainExpression(string))

        if 'omega' in symbols:
            if 't' in symbols:
                # Handle cos(omega * t)
                return self.add(TimeDomainExpression(string))
            return self.add(AngularFourierDomainExpression(string))

        if 'f' in symbols:
            # TODO, handle AC of different frequency
            return self.add(FourierDomainExpression(string))

        return self.add(TimeDomainExpression(string))

    def _add_noise(self, value):

        if value.nid not in self:
            self[value.nid] = value
        else:
            self[value.nid] += value
            if self[value.nid] == 0:
                self.pop(value.nid)

    def add(self, value):
        """Add a value into the superposition."""

        try:
            val = value.expr
        except:
            val = value

        # Avoid triggering __eq__ for Super otherwise have infinite recursion
        if not isinstance(value, Superposition):
            if val == 0:
                return

        if hasattr(self, '_decomposition'):
            delattr(self, '_decomposition')

        if isinstance(value, Superposition):
            for kind, value1 in value.items():
                self.add(value1)
            return

        if isinstance(value, str):
            return self._parse(value)

        if isinstance(value, NoiseExpression):
            self._add_noise(value)
            return

        value = expr(value)
        if value.has(sym.Float):
            # Avoid 8.00000000000, etc.
            value.expr = sym.nsimplify(value.expr)

        kind = self._kind(value)
        if kind is None:
            if self.is_voltage:
                value = value.as_voltage()
            else:
                value = value.as_current()
            kind = self._kind(value)

        if kind is None:
            raise ValueError('Cannot handle value %s of type %s' %
                             (value, type(value).__name__))

        if kind not in self:
            self[kind] = value
        else:
            self[kind] += value

    @property
    def magnitude(self):
        """Return the magnitude if phasor."""

        phasor = self.phasor()
        return phasor.magnitude

    @property
    def abs(self):
        """Return the magnitude if phasor."""
        return self.magnitude

    @property
    def omega(self):
        """Return angular frequency if phasor."""
        return self.omega

    @property
    def phase(self):
        """Return the phase if phasor."""

        phasor = self.phasor()
        return phasor.phase

    @property
    def dc(self):
        """Return the DC component."""
        return self.select('dc')

    @property
    def _single(self):
        """Extract single expression; raise error if a superposition."""

        decomp = self.decompose()

        if len(decomp) > 1:
            raise ValueError("""
This is superposition of multiple signals.  You will need to extract a single
expression.""")

        key = list(decomp)[0]
        return decomp[key]

    def phasor(self, **kwargs):
        """Return phasor if have a single AC component otherwise raise error."""

        expr1 = self._single
        return expr1.phasor(**kwargs)

    @property
    def ac(self):
        """Return the AC components."""
        if 't' in self.keys():
            self = self.decompose()
        return ExprDict({k: v for k, v in self.items() if k in self.ac_keys()})

    @property
    def transient(self):
        """Return the transient component."""
        return self.select('s').time()

    @property
    def s(self):
        """Return the s-domain representation of the transient component.
        This is not the full s-domain representation as returned by the
        laplace method V.laplace() or V(s).

        This attribute may be deprecated due to possible confusion."""
        return self.select('s')

    @property
    def n(self):
        result = self.decompose_to_domain(0, 'n')
        for key in self.noise_keys():
            result += self[key]
        return result

    @property
    def noise(self):
        """Return the total noise."""
        return self.n

    def time(self, **assumptions):
        """Convert to time domain."""

        result = domain_quantity_to_class('time', quantity=self.quantity)(0)

        # TODO, integrate noise
        for val in self.values():
            result += val.time(**assumptions)
        return result

    def transient_response(self, tvector=None):
        """Evaluate transient (impulse) response."""

        texpr = self.time()

        if tvector is None:
            return texpr

        return texpr.evaluate(tvector)

    def frequency_response(self, fvector=None):
        """Convert to frequency domain and evaluate response if frequency
        vector specified.

        """

        X = self.fourier()

        if fvector is None:
            return X

        return X.evaluate(fvector)

    def laplace(self, **assumptions):
        """Convert to s-domain."""

        result = domain_quantity_to_class('laplace', quantity=self.quantity)(0)
        result.is_causal = True
        for val in self.values():
            result += val.laplace(**assumptions)
        return result

    def fourier(self, **assumptions):
        """Convert to Fourier domain."""

        result = domain_quantity_to_class('fourier', quantity=self.quantity)(0)
        result.is_causal = True
        for val in self.values():
            result += val.fourier(**assumptions)
        return result

    def angular_fourier(self, **assumptions):
        """Convert to angular Fourier domain."""

        # TODO, could optimise.
        tresult = self.time(**assumptions)
        return tresult.angular_fourier(**assumptions)

    def canonical(self):
        new = self.__class__()
        for kind, value in self.items():
            new[kind] = value.canonical()

        return new

    def simplify(self):
        new = self.__class__()
        for kind, value in self.items():
            new[kind] = value.simplify()

        return new

    def oneport(self):
        """Create oneport component."""
        return self.cpt()

    def rms(self):
        return self.n.rms()

    def asd(self):
        return self.n.asd()

    def psd(self):
        return self.n.psd()

    def force_time(self):

        if state.force_time:
            return self.time()
        return self

    def limit(self, var, value, dir='+'):
        """Determine limit of expression(var) at var = value.
        If `dir == '+'` search from right else if `dir == '-'`
        search from left."""

        new = self.__class__()
        for kind, val in self.items():
            new[kind] = val.limit(var, value, dir)

        return new


from .cexpr import ConstantDomainExpression  # nopep8
from .fexpr import FourierDomainExpression  # nopep8
from .sexpr import LaplaceDomainExpression  # nopep8
from .texpr import TimeDomainExpression  # nopep8
from .noiseexpr import NoiseExpression  # nopep8
from .phasor import PhasorDomainExpression  # nopep8
from .omegaexpr import AngularFourierDomainExpression  # nopep8
from .symbols import s, omega  # nopep8

# See expr.py

try:
    from .printing import latex
    formatter = sys.displayhook.shell.display_formatter.formatters['text/latex']

    formatter.type_printers[Superposition] = Superposition._repr_latex_
except:
    pass



"""This module provides the ComponentNamer class for automatic
naming of components.

Copyright 2020--2023 Michael Hayes, UCECE

"""


class ComponentNamer(object):

    def __init__(self):

        self.names = []

    def name(self, cpt_type, names):

        m = 1
        while True:
            name = cpt_type + str(m)
            if name not in names and name not in self.names:
                self.names.append(name)
                return name
            m += 1

    def __call__(self, cpt_type, names=None):
        """Create component name based on `cpt_type` (`R`, `L`, etc.)
        and the name_kind (`anon`, `t`, etc).

        `names` is a list of known names to avoid."""

        return self.name(cpt_type, names)



from lcapy import (symbol, texpr, cos, sin, exp,
                   expr, pi, rect, sinc, tri, trap, delta, sign, H, j, t, f)

alpha = symbol('alpha')
t0 = symbol('t0')
f0 = symbol('f0')
w0 = 2 * pi * f0

sigs = [texpr('x(t)'), texpr('x(a * t)'), texpr('x(t - tau)'),
        cos(w0 * t), sin(w0 * t), exp(j * w0 * t),
        texpr(1), t, t**2, 1 / t, delta(t), delta(t - t0),
        H(t), t * H(t), sign(t),
        rect(t), sinc(t), tri(t), trap(t, alpha),
        exp(-abs(t)), exp(-t) * H(t)]

for sig in sigs:
    print(':math:`%s \\longleftrightarrow %s`\n' %
          (sig.latex(), sig(f).latex()))



from matplotlib.pyplot import savefig, subplots
from numpy import linspace
from lcapy import *

figs, axes = subplots(1)
cos(2 * t).plot(axes=axes, label='cos')
sin(2 * t).plot(axes=axes, label='sin')
axes.legend()

savefig(__file__.replace('.py', '.png'), bbox_inches='tight')




"""This module performs transformations between domains.

Copyright 2018--2022 Michael Hayes, UCECE

"""

from .impedance import impedance  # nope8p
from .sym import sympify, pi
from .sym import dt
from .symbols import f, s, t, omega, j, jf, jw, jw0, Omega, F, k, n, z
from .symbols import domain_vars, domain_var_ids
from .nexpr import n
from .expr import expr as expr1
from .expr import Expr


def transform(expr, arg, **assumptions):
    """If arg is a domain variable perform domain transformation,
    otherwise perform substitution.

    Note (1 / s)(omega) will fail since 1 / s is assumed not to be
    causal and so the Fourier transform is unknown.  However,
    impedance(1 / s)(omega) will work since an impedance is assumed to
    be causal.  Alternatively, use (1 / s)(omega, causal=True).

    Transforming from s->jomega is fast since it just requires
    a substitution of s with jomega.

    Transforming from s->omega, s->Omega, s->f, or s->F can be slow since
    this requires a inverse Laplace transform followed by a Fourier
    transform.  However, if the expression is causal and the
    expression is lossy when s is replaced by jw, the result can be
    found by substituting jw or 2 * 2 * pi * f for s.  This does not
    apply for an expression such as Z = 1 / (s * C).

    """

    arg = expr1(arg)

    # Handle things like (3 * s)(5 * s)
    if isinstance(expr, arg.__class__) and not isinstance(expr, Superposition):
        return expr.subs(arg)

    # Handle expr(t), expr(s), expr(f), expr(omega), expr(jomega), expr(3j)
    try:
        if arg is t:
            return expr.time(**assumptions)
        elif arg is s:
            return expr.laplace(**assumptions)
        elif arg is f:
            return expr.fourier(**assumptions)
        elif arg is omega:
            return expr.angular_fourier(**assumptions)
        elif arg is Omega:
            return expr.norm_angular_fourier(**assumptions)
        elif arg is F:
            return expr.norm_fourier(**assumptions)
        elif arg is jf:
            return expr.frequency_response(**assumptions)
        elif arg is jw:
            return expr.angular_frequency_response(**assumptions)
        elif arg.has(j):
            # Note, substituting s = jw is valid for the phasor domain
            # but does not always produce the freqguency response since
            # the evaluation point can be outside the region of
            # convergence.  For example, the LT for u(t), t * u(t),
            # and exp(a * t) for positive a, i.e., the impulse
            # responses for marginally stable and unstable systems.
            # For the frequency response use omega.
            return expr.phasor_ratio(omega=arg / j, **assumptions)
        elif arg is n:
            return expr.discrete_time(**assumptions)
        elif arg is k:
            return expr.discrete_frequency(**assumptions)
        elif arg is z:
            return expr.zdomain(**assumptions)
    except AttributeError:
        raise AttributeError('Transform of %s domain to %s domain not implemented' % (
            expr.domain, arg.domain))

    # Handle expr(texpr), expr(sexpr), expr(fexpr), expr(omegaexpr).
    # For example, expr(2 * f).
    result = None
    if isinstance(arg, TimeDomainExpression):
        result = expr.time(**assumptions)
    elif isinstance(arg, LaplaceDomainExpression):
        result = expr.laplace(**assumptions)
    elif isinstance(arg, FourierDomainExpression):
        result = expr.fourier(**assumptions)
    elif isinstance(arg, NormFourierDomainExpression):
        result = expr.norm_fourier(**assumptions)
    elif isinstance(arg, AngularFourierDomainExpression):
        result = expr.angular_fourier(**assumptions)
    elif isinstance(arg, NormAngularFourierDomainExpression):
        result = expr.norm_angular_fourier(**assumptions)
    elif isinstance(arg, AngularFrequencyResponseDomainExpression):
        result = expr.angular_frequency(**assumptions)
    elif arg.has(j):
        result = expr.phasor(omega=arg / j, **assumptions)
    elif isinstance(arg, DiscreteTimeDomainExpression):
        result = expr.discrete_time(**assumptions)
    elif isinstance(arg, DiscreteFourierDomainExpression):
        result = expr.discrete_frequency(**assumptions)
    elif isinstance(arg, ZDomainExpression):
        result = expr.zdomain(**assumptions)
    elif arg.is_constant:
        if not isinstance(expr, Superposition):
            result = expr.time(**assumptions)
        else:
            result = expr
    else:
        raise ValueError(
            'Can only return t, f, s, F, omega, Omega, or jw domains')

    return result.subs(arg, **assumptions)


def call(expr, arg, **assumptions):

    from numpy import ndarray, array

    if isinstance(arg, (tuple, list)):
        return [expr._subs1(expr.var, arg1) for arg1 in arg]

    elif isinstance(arg, ndarray):
        return array([expr._subs1(expr.var, arg1) for arg1 in arg])

    arg = expr1(arg)
    if arg.is_constant:
        return expr.subs(arg)

    return expr.transform(arg, **assumptions)


def select(expr, kind):

    if not isinstance(kind, str):
        return expr.subs(j * kind)

    # If kind is an expr, then will add 'dc', 'time', etc. as symbols.

    if kind == 't':
        return expr.time()
    elif kind in ('dc', 'time'):
        return expr.subs(0)
    elif kind in ('s', 'ivp', 'super', 'laplace'):
        return expr.laplace()
    elif kind == 'f':
        return expr.fourier()
    elif kind == 'omega':
        return expr.angular_fourier()
    elif kind == 'Omega':
        return expr.norm_angular_fourier()
    elif kind == 'F':
        return expr.norm_fourier()
    elif isinstance(kind, str) and kind.startswith('n'):
        return expr.angular_fourier()
    else:
        raise RuntimeError('unknown kind')


from .fexpr import FourierDomainExpression  # nopep8
from .sexpr import LaplaceDomainExpression  # nopep8
from .texpr import TimeDomainExpression  # nopep8
from .omegaexpr import AngularFourierDomainExpression  # nopep8
from .normfexpr import NormFourierDomainExpression  # nopep8
from .normomegaexpr import NormAngularFourierDomainExpression  # nopep8
from .jomegaexpr import AngularFrequencyResponseDomainExpression  # nopep8
from .nexpr import DiscreteTimeDomainExpression  # nopep8
from .kexpr import DiscreteFourierDomainExpression  # nopep8
from .zexpr import ZDomainExpression  # nopep8
from .super import Superposition  # nopep8
from .current import current  # nopep8
from .voltage import voltage  # nopep8
from .admittance import admittance  # nopep8
from .transfer import transfer  # nopep8



from matplotlib.pyplot import savefig
from lcapy import Circuit, f

a = Circuit('filter1.sch')
H = a.transfer('P1', 'P2')
Hv = H.subs({'R1':22, 'C1':100e-9, 'R2':1e6, 'C2':1e-9})
Hv.bode_plot((0.1, 10e3))

savefig(__file__.replace('.py', '.png'), bbox_inches='tight')



from lcapy import *
from numpy import logspace
from matplotlib.pyplot import savefig

N = R(10) + C(1e-4) + L(1e-3)

vf = logspace(0, 5, 400)
N.Z(j2pif).magnitude.plot(vf)

savefig('series-RLC3-Z.png')



"""This module provides capacitance support.

Copyright 2021 Michael Hayes, UCECE

"""
from .cexpr import cexpr
from .units import u as uu
from warnings import warn


def capacitance(arg, **assumptions):

    expr1 = cexpr(arg, frequency=True, **assumptions)
    if expr1.is_imaginary:
        warn('Capacitance %s should be real' % expr1)

    expr1.units = uu.farad
    return expr1



"""This module provides the State class that maintains the global state,
such as the symbol context.

Copyright 2014--2022 Michael Hayes, UCECE

"""

from sympy.assumptions.assume import global_assumptions

from .context import Context
from .config import check_units, abbreviate_units, loose_units, show_units
from .config import canonical_units, printing_order
from .printing_config import PrintingConfig
from .symbolregistry import SymbolRegistry
from copy import copy
from warnings import warn


class State(object):
    """This maintains Lcapy's state, such as the defined symbols, and
    behaviour.

    `loose_units` (default True) allows constants to be added to quantities
    `show_units` (default False) prints the units after an expression
    `canonical_units` (default True) converts units to canonical form, e.g., V / A is shown as ohms.
    'printing.abbreviate_units` (default True) prints V rather than volts
    'printing.order` (default `none`) controls the order that symbols in an expression are printed.

    """

    def __init__(self):
        self.symbols = SymbolRegistry()
        self.context = Context()
        self.context.symbols = self.symbols
        self.previous_context = []
        self.printing = PrintingConfig()

        # With loose_units can add constants to quantities, e.g., voltage(1) + 2
        # or impedance(2) == 2.
        self.loose_units = loose_units
        self.check_units = check_units
        self.show_units = show_units
        self.canonical_units = canonical_units
        self.printing.abbreviate_units = abbreviate_units
        self.printing.order = printing_order

        self.warn_subs = False
        self.break_subs = False
        self.notify_symbol_add = False
        self.warn_unknown_symbol = False
        self.break_unknown_symbol = False

        self.zero_initial_conditions = False

        # The following can either be `allow`, `warn`, or `error`.
        # They are used in expr.c:__mul__ to validate multiplications.
        self.s_times_t = 'warn'
        self.t_times_s = 'warn'
        self.f_times_t = 'warn'
        self.t_times_f = 'warn'
        self.w_times_t = 'warn'
        self.t_times_w = 'warn'

        # The following can either be `allow`, `warn`, or `error`.
        # They are used in the expression class constructors.
        self.f_in_t = 'warn'
        self.s_in_t = 'warn'
        self.w_in_t = 'warn'

        self.f_in_s = 'warn'
        self.t_in_s = 'warn'
        self.w_in_s = 'warn'

        self.s_in_f = 'warn'
        self.t_in_f = 'warn'
        self.w_in_f = 'warn'

        self.f_in_w = 'warn'
        self.s_in_w = 'warn'
        self.t_in_w = 'warn'

        self.f_in_jf = 'warn'
        self.s_in_jf = 'warn'
        self.t_in_jf = 'warn'

        self.f_in_jw = 'warn'
        self.s_in_jw = 'warn'
        self.t_in_jw = 'warn'

        self.pdb_on_warn = False
        self.error_on_warn = False

        # 'passive', 'active', 'hybrid', None for default ('hybrid')
        self.current_sign_convention = None

    def new_context(self):

        context = Context()
        context.symbols = self.symbols
        return context

    def switch_context(self, context):

        self.previous_context.append(self.context)
        self.context = context

        global_assumptions.clear()
        global_assumptions.update(self.context.assumptions)

    def restore_context(self):

        self.context.assumptions.update(global_assumptions)

        self.context = self.previous_context.pop()

        global_assumptions.clear()
        global_assumptions.update(self.context.assumptions)

    @property
    def abbreviate_units(self):

        return self.printing.abbreviate_units

    @abbreviate_units.setter
    def abbreviate_units(self, val):
        """Enable/disable printing of units in abbreviated form."""

        self.printing.abbreviate_units = val


def validate(value, message):

    if value == 'allow':
        pass
    elif value == 'warn':
        if state.pdb_on_warn:
            import pdb
            pdb.set_trace()
        if state.error_on_warn:
            raise ValueError(message)
        else:
            warn(message)
    elif value == 'error':
        raise ValueError(message)
    else:
        raise RuntimeError('Unexpected option ' + value)


state = State()



from lcapy import *
from numpy import linspace
from matplotlib.pyplot import savefig

N = Vstep(20) + R(10) + L(1e-2, 0)

vt = linspace(0, 0.01, 1000)
N.Isc(t).plot(vt)

savefig('series-VRL1-isc.png')



"""This module provides the Fitter class.  This uses optimization
techniques to find the parameters of an expression that best fits measured data.

Copyright 2022--2023 Michael Hayes, UCECE

"""

from scipy.optimize import brute, fmin, curve_fit, minimize
from numpy import iscomplexobj, hstack, zeros, sqrt


class FitterResult(object):

    def __init__(self, params, rmse, expr):

        self.params = params
        self.rmse = rmse
        self.expr = expr.subs(params)

    def __call__(self, arg):

        return self.expr.evaluate(arg)


class Fitter(object):
    """Data fitting."""

    def __init__(self, expr):

        self.expr = expr
        self.symbols = expr.symbols
        self.symbols.pop(str(expr.var))

    def _make_defs(self, params, ranges):

        defs = {}
        for m, r in enumerate(ranges.items()):
            defs[r[0]] = params[m]
        return defs

    def _make_ranges(self, ranges, positive, param_min=None, param_max=None):
        """Create missing ranges."""

        if ranges is None:
            ranges = {}

        if param_min is None:
            param_min = 0 if positive else -1e9
        if param_max is None:
            param_max = 1e9

        for symbol in self.symbols:
            if symbol not in ranges:
                # Perhaps make (-inf, inf) for unbounded
                # but will need to fix initial guess.
                ranges[symbol] = (param_min, param_max)
        return ranges

    def model(self, params, x, ranges):

        defs = self._make_defs(params, ranges)

        return self.expr.subs(defs).evaluate(x)

    def _optimize_brute(self, x, y, ranges=None, Ns=10, finish='fmin',
                        **kwargs):

        kwargs.pop('method', None)
        self.verbose = kwargs.pop('verbose', 0)

        if finish in ('none', 'None', ''):
            finish = None
        elif finish == 'fmin':
            finish = fmin

        oranges = []
        for k, r in ranges.items():
            if len(r) == 2:
                # Note, a complex value specifies the number of steps
                # (see numpy.mgrid).
                oranges.append(slice(r[0], r[1], complex(Ns)))
            elif len(r) == 3:
                oranges.append(slice(r[0], r[1], complex(r[2])))
            else:
                raise ValueError('Range %s can only have 2 or 3 values' % r)

        iscomplex = iscomplexobj(y)

        def func(params):

            yp = self.model(params, x, ranges)
            if iscomplex:
                return (abs(y - yp)**2).mean()

            return ((y - yp)**2).mean()

        params, rmse, foo, bar = brute(func, oranges, Ns=Ns, args=(),
                                       finish=finish, full_output=1)

        defs = self._make_defs(params, ranges)
        return FitterResult(defs, rmse, self.expr)

    def _optimize_curvefit(self, x, y, ranges=None, method='trf',
                           ftol=1e-14, xtol=1e-14, maxfev=1e5,
                           **kwargs):

        kwargs.pop('Ns', None)
        kwargs.pop('finish', None)

        bounds_min = zeros(len(ranges))
        bounds_max = zeros(len(ranges))

        for m, r in enumerate(ranges.values()):
            if len(r) in (2, 3):
                bounds_min[m] = r[0]
                bounds_max[m] = r[1]
            else:
                raise ValueError('Range %s can only have 2 or 3 values' % r)

        bounds = (bounds_min, bounds_max)

        # Initial guess.
        p0 = 0.5 * (bounds_min + bounds_max)

        iscomplex = iscomplexobj(y)

        if iscomplex:
            y = hstack((y.real, y.imag))

        def func(x, *params):

            yp = self.model(params, x, ranges)
            if iscomplex:
                yp = hstack((yp.real, yp.imag))
            return yp

        params, cov = curve_fit(func, x, y, p0=p0,
                                bounds=bounds, ftol=ftol, xtol=xtol,
                                maxfev=maxfev, **kwargs)

        yp = func(x, *params)
        rmse = ((y - yp)**2).mean()

        defs = self._make_defs(params, ranges)
        return FitterResult(defs, rmse, self.expr)

    def _optimize_minimize(self, x, y, ranges=None,
                           method='Nelder-Mead', **kwargs):

        kwargs.pop('Ns', None)
        kwargs.pop('finish', None)

        bounds = []
        p0 = []

        for m, r in enumerate(ranges.values()):
            if len(r) in (2, 3):
                bounds.append((r[0], r[1]))
                p0.append(0.5 * (r[0] + r[1]))
            else:
                raise ValueError('Range %s can only have 2 or 3 values' % r)

        iscomplex = iscomplexobj(y)

        if iscomplex:
            y = hstack((y.real, y.imag))

        def func(params):

            yp = self.model(params, x, ranges)

            if iscomplex:
                return (abs(y - yp)**2).mean()
            return ((y - yp)**2).mean()

        results = minimize(func, x0=p0, bounds=bounds, method=method, **kwargs)

        params = results.x

        yp = func(params)
        rmse = ((y - yp)**2).mean()

        defs = self._make_defs(params, ranges)
        return FitterResult(defs, rmse, self.expr)

    def _optimize1(self, x, y, ranges=None, method='trf', **kwargs):

        if method == 'brute':
            return self._optimize_brute(x, y, ranges, **kwargs)
        elif method in ('lm', 'trf', 'dogbox'):
            return self._optimize_curvefit(x, y, ranges, method, **kwargs)
        elif method in ('Nelder-Mead', 'Powell'):
            return self._optimize_minimize(x, y, ranges, method, **kwargs)
        else:
            return self._optimize_minimize(x, y, ranges, method, **kwargs)

    def optimize(self, x, y, ranges=None, method='trf', iterations=1,
                 positive=False, param_min=None, param_max=None, **kwargs):

        ranges = self._make_ranges(ranges, positive, param_min, param_max)

        iscomplex = iscomplexobj(y)

        if iterations <= 1:
            return self._optimize1(x, y, ranges, method, **kwargs)

        tol = kwargs.get('tol', 1e-6)

        rmse_prev = None
        x1 = x
        y1 = y
        for i in range(iterations):

            result = self._optimize1(x1, y1, ranges, method, **kwargs)

            params = result.params
            yp = self.expr.subs(params).evaluate(x)

            e = abs(y - yp)
            sd = e.std()
            m = e <= 3 * sd
            num_outliers = len(x) - sum(m)
            if num_outliers == 0:
                break

            x1 = x[m]
            y1 = y[m]

            if len(x1) == 0:
                raise ValueError('Outlier removal failed')

            e = abs(y1 - yp[m])
            rmse = sqrt((e**2).mean())

            if False:
                print('Removed %d outliers at iteration %d: sd=%f, rmse=%f' %
                      (num_outliers, i + 1, sd, rmse))

            result.rmse = rmse
            if rmse < tol:
                break
            if rmse_prev is not None and rmse == rmse_prev:
                break
            rmse_prev = rmse

        return result


def fit(expr, x, y, ranges, method='trf', Ns=10, **kwargs):

    return Fitter(expr).optimize(x, y, ranges=ranges, method=method,
                                 Ns=Ns, **kwargs)


def test1():

    from lcapy import expr, f
    from numpy import arange

    e = expr('a * exp(-t  / tau) * u(t)')

    tv = arange(100)

    vv = e.subs({'a': 1, 'tau': 10}).evaluate(tv)

    results = fit(e, tv, vv, method='brute',
                  ranges={'a': (0, 10), 'tau': (1, 20)})
    print(results.params)

    results = fit(e, tv, vv, method='trf',
                  ranges={'a': (0, 10), 'tau': (1, 20)})
    print(results.params)

    E = e(f)

    fv = arange(100)

    Vv = E.subs({'a': 1, 'tau': 10}).evaluate(fv)

    results = fit(E, fv, Vv, method='brute',
                  ranges={'a': (0, 10), 'tau': (1, 20)})
    print(results.params)

    results = fit(E, fv, Vv, method='trf',
                  ranges={'a': (0, 10), 'tau': (1, 20)})
    print(results.params)
